import unittest
from activedetect.loaders.csv_loader import CSVLoader

class TestLoad(unittest.TestCase):
	
	def testCSVLoader(self):
		"""
		Tests the CSV Loading primatives
		"""
		c = CSVLoader()
		l = c.loadFile('test/resources/f1.csv')
		assert(c.delim == (',', '"'))

		c = CSVLoader()
                l = c.loadFile('test/resources/f2.csv')
		assert(c.delim == (':', '"'))

		c = CSVLoader()
                l = c.loadFile('test/resources/f3.csv')
		assert(c.delim == (',', '"'))
		assert(len(l[0]) == 2)

		c = CSVLoader()
		l = c.loadFile('test/resources/f4.csv')
		assert(c.delim == (',', '"'))

		c = CSVLoader()
                l = c.loadFile('test/resources/f5.csv')
                assert(c.delim == ('\t', '"'))

if __name__ == '__main__':
    unittest.main()




"""
Detects Semantic Errors in A Dataset Using 
Word2Vec
"""
from gensim.models.word2vec import Word2Vec
from gensim.models import word2vec
import numpy as np
import re
from ErrorModule import ErrorModule

class SemanticErrorModule(ErrorModule):

	def __init__(self, corpus='corpora/text8', thresh=3.5, fail_thresh=0.8):
		sentences = word2vec.Text8Corpus(corpus)
		
		self.model = word2vec.Word2Vec(sentences)

		self.thresh = thresh

		self.fail_thresh = fail_thresh


	"""
	Returns the subset of a domain that is potentially
	erroneous

	use dictionary instead

	O(N^2): N distinct values
	"""
	def predict(self, vals):
		
		domain = set(vals)

		error = []
		incorpus = []
		full_corpus = []
		
		#for each val in the domain
		for d in domain:

			#tokenizes categorical attrs
			tokens = re.findall(r"[\w']+", d)

			#if no tokens error
			if len(tokens) == 0:
				error.append(d)
			else:

				#iterate through tokens add to corpus
				match = False
				for t in tokens:
					if t in self.model:
						incorpus.append(t)
						full_corpus.append(t)
						match = True

				#if no matches error
				if not match:
					error.append(d)


		#if there are too few matches just return
		if (len(error)+0.0)/len(domain) > self.fail_thresh:
			return [], full_corpus


		#build similarity graph, take in-degree
		aggsim = {}
		vals = []
		for i in incorpus:
			agg = 0
			for j in incorpus:
				agg = agg + self.model.similarity(i,j)

			aggsim[i] = agg
			vals.append(agg)


		#take MAD to filter corpus
		mad = self.mad(vals)
		median = np.median(vals)

		for a in aggsim:
			if np.abs(aggsim[a] - median) > self.thresh*mad:
				error.append(a)
				incorpus.remove(a)



		#return error and incorpus
		return error, incorpus
		


	""" Median Absolute Deviation: a "Robust" version of standard deviation.
    Indices variabililty of the sample.
    https://en.wikipedia.org/wiki/Median_absolute_deviation 
    """
	def mad(self, arr):
		arr = np.ma.array(arr).compressed() # should be faster to not use masked arrays.
		med = np.median(arr)
		return np.median(np.abs(arr - med))


	"""
	Turns an error set into a set of records
	"""
	def getRecordSet(self, errors, dataset, col):

		indices = []
		erecords = []

		for i,d in enumerate(dataset):
			val = d[col]
			tokens = set(re.findall(r"[\w']+", val))

			match = False

			for e in errors:

				if e in tokens or \
				   (len(tokens) == 0 and e in val):
					match = True
					break

			if match:
				indices.append(i)
				erecords.append(d)

		return erecords, indices

	"""
	"""
	def desc(self):
		return "A value was found with an abnormal word2vec similarity score thresh= " + str(self.thresh)


	def availTypes(self):
		return ['categorical']






		



#!/usr/bin/env python
import csv
import numpy as np
from sklearn.preprocessing import FunctionTransformer
from sklearn.feature_extraction.text import CountVectorizer
import scipy
import unicodedata

#basic error handling
def tryParse(X):

	vals = []

	if X.shape == (1,1):
		try:
			vals.append(float(X.tolist()[0][0]))
		except ValueError:
			vals.append(0)

		return vals


	for x in np.squeeze(X.T):
		try:
			vals.append(float(x))
		except ValueError:
			vals.append(0)

	return vals

#converts the labeled dataset into features and labels
def featurize(features_dataset, types):
	feature_list = []

	for i,t in enumerate(types):
		col = [f[i] for f in features_dataset]

		if t == "string" or t == "categorical" or t =="address":
			vectorizer = CountVectorizer(min_df=1, token_pattern='\S+')
			feature_list.append(vectorizer.fit_transform(col))
		else:
			vectorizer = FunctionTransformer(tryParse)
			feature_list.append(scipy.sparse.csr_matrix(vectorizer.fit_transform(col)).T)

	features = scipy.sparse.hstack(feature_list).tocsr()
	return features



"""
Detects Semantic Errors in A Dataset Using 
Word2Vec
"""
from gensim.models.word2vec import Word2Vec
from gensim.models import word2vec
import numpy as np
import re
from ErrorModule import ErrorModule

class StringSimilarityErrorModule(ErrorModule):


	#takes in a list of words
	def __init__(self, thresh=3.5):		
		self.thresh = thresh


	"""
	Returns the subset of a records that are potentially
	erroneous

	use dictionary instead

	O(N^2): N distinct values
	"""
	def predict(self, strings):
		
		self.model = word2vec.Word2Vec(strings,hs=1)

		error = []
		incorpus = []

		incorpus = strings
		error = []
		
		#for each val in the column
		string_scores = []
		scoredict = {}

		for s in strings:

			cleaned_string = s.lower().strip()

			if len(cleaned_string) == 0:
				error.append(s)
				incorpus.remove(s)
			else:
				score = np.squeeze(self.model.score([cleaned_string]))/len(cleaned_string)
				string_scores.append(score)
				scoredict[s] = score

		mad = np.std(string_scores)
		median = np.mean(string_scores)

		for s in scoredict:
			if np.abs(scoredict[s] - median) > self.thresh*mad:
				error.append(s)
				incorpus.remove(s)

		return error, incorpus
		


	""" Median Absolute Deviation: a "Robust" version of standard deviation.
    Indices variabililty of the sample.
    https://en.wikipedia.org/wiki/Median_absolute_deviation 
    """
	def mad(self, arr):
		arr = np.ma.array(arr).compressed() # should be faster to not use masked arrays.
		med = np.median(arr)
		return np.median(np.abs(arr - med))


	"""
	Turns an error set into a set of records
	"""
	def getRecordSet(self, errors, dataset, col):

		indices = []
		erecords = []

		for i,d in enumerate(dataset):

			val = d[col]

			if val in set(errors):
				indices.append(i)
				erecords.append(d)
				
		return erecords, indices

	def desc(self):
		return "A string was found that was not well predicted by a sequential module thresh= " + str(self.thresh)


	def availTypes(self):
		return ['string']




		



import csv
import numpy as np

class CSVLoader:
	"""
	This class provides a wrapper to load csv files into the system
	"""

	def __init__(self, delimiter=None, quotechar=None):
		"""
		You can create a CSV loader with a specified delimiter
		or quote character. Or it will just try them all.
		"""
		self.DELIMITERS = [',', '\t', ':', '~', '|']
		self.QUOTECHAR = ['"', "'", '|', ':']

		if delimiter != None:
			self.DELIMITERS = [delimiter]

		if quotechar != None:
			self.QUOTECHAR = [quotechar]


	def __load(self, fname, delimiter, quotechar):
		"""
		Internal method to load a CSVfile given a delimiter and 
		quote character
		"""
		with open(fname,'rb') as file:
			try:
				return [r for r in csv.reader(file, 
							  delimiter=delimiter, 
							  quotechar=quotechar)]
			except:
				return None

		return None


	def __score(self,parsed_file):
		"""
		This method assigns a score to all of the parsed files.
		We count the variance in the row length
		"""
		rowstd = np.std([len(row) for row in parsed_file])

		#catch degenerate case
		if rowstd == 0 and len(parsed_file[0]) == 1:
			return float("inf")
		else:
			return rowstd


	def loadFile(self, fname):
		"""
		External method to load a file
		"""
		parsed_files = []

		#try out all of the options in the delimiter and quotechar set
		for delimiter in self.DELIMITERS:
			for quotechar in self.QUOTECHAR:
				loaded = self.__load(fname, delimiter, quotechar)
				if not loaded == None:
					parsed_files.append(((delimiter,quotechar), loaded))

		#score each of the parsed files
		scored_parses = [(self.__score(p[1]), p[0], p[1]) for p in parsed_files]
		scored_parses.sort()

		#print [(s[0],s[1]) for s in scored_parses]

		self.delim = scored_parses[0][1]

		return scored_parses[0][2]






"""
Detects Quantitative Errors in a dataset using
standard deviations
"""
import numpy as np
from ErrorModule import ErrorModule

class QuantitativeErrorModule(ErrorModule):

	def __init__(self, thresh=3.5):
		self.thresh = thresh


	"""
	Returns the subset of a domain that is potentially
	erroneous
	"""
	def predict(self, vals):
		
		vals = filter( lambda x : not np.isinf(x) ,[self.tryParse(v) for v in list(vals)])

		std = np.std(vals)
		mean = np.mean(vals)
		
		#make a copy
		incorpus = []
		error = []
		incorpus.extend(vals)

		for a in vals:
			if np.abs(a - mean) > self.thresh*std:
				error.append(a)
				incorpus.remove(a)

		return error, incorpus


	def tryParse(self, num):
		try:
			return float(num)
		except ValueError:
			return np.inf


	"""
	Turns an error set into a set of records
	"""
	def getRecordSet(self, errors, dataset, col):

		indices = []
		erecords = []

		for i,d in enumerate(dataset):
			val = self.tryParse(d[col])

			match = False

			for e in errors:

				if e == val:
					match = True

			#match or can't parse
			if match or np.isinf(val):
				indices.append(i)
				erecords.append(d)

		return erecords, indices

	"""
	Returns a description
	"""
	def desc(self):
		return "A numerical value was found with a value of greater than > "+ str(self.thresh) + " stds from the mean"

	"""
	Returns where it is applicable
	"""
	def availTypes(self):
		return ['numerical']